
#include "s3mfun.h"

#include <map>
#include <string>
#include <vector>
#include <utility>
#include <stdio.h>

using std::map;
using std::string;
using std::vector;
using std::make_pair;

#define SOUND_DECODE_LENGTH 16
#define PLAYBACK_RATE 44100
//#define PLAYBACK_RATE so.playback_rate

#ifdef SPClog
/*** Functions for saving the soundtrack ***/
static class S3Mmessage
{
	unsigned S3Mclock[8], S3Mclock2[8], S3Mprevtick[8];
	bool S3Mkonpending[8];
	double S3Mkontime[8];
	FILE *fp;
	
    double GetClock(unsigned ch)
    {
        double time = (S3Mclock[ch] + S3Mclock2[ch]) / (double)PLAYBACK_RATE;
        return time;
    }
    void FlushKon(unsigned ch)
    {
        if(S3Mkonpending[ch])
        {
            /* Delay key-ons until freq is known */
            S3Mkonpending[ch] = false;
            /* Enters the attack phase */
            fprintf(fp, "O%u@%g #kon at %g\n", ch, GetClock(ch), S3Mkontime[ch]);
        }
    }
public:
	S3Mmessage()
	{
	    if(!fp) fp = fopen("spcplay.log", "wt");
	}
	~S3Mmessage()
	{
	    fclose(fp);
	}

    inline void SampleData(unsigned id, const vector<signed short>& data)
    {
        /* Records a sample */
        char Buf[64];
        sprintf(Buf, "sample%X.raw", id);
        FILE *sfp = fopen(Buf, "wb");
		for(unsigned a=0; a<data.size(); ++a)
		{
			signed short sample = data[a];
			fputc(sample&255, sfp);
			fputc(sample>>8, sfp);
		}
        fclose(sfp);
        fprintf(fp, "D%X:%s #Sample %X data file\n", id, Buf, id);
    }
    inline void SampleInfo(unsigned id, unsigned loopbeg, unsigned loopend, int state)
    {
        /* Describes a sample */
        fprintf(fp, "S%X:%u-%u %d #Sample %X: Loop %u-%u, %s\n",
            id,loopbeg,loopend,state,
            id, loopbeg, loopend,
            state==3 ? "complete, looping"
          : state==2 ? "looping"
          : state==1 ? "complete"
          : state==0 ? "partial"
          : "unknown");
    }
    inline void InstrInfo(unsigned id, int a,int d,int s,int r,int sl,unsigned smpid)
    {
        /* Describes an instrument */
        fprintf(fp, "I%u:%d %d %d %d %d %X  # Instrument %u ADSR,suslev,sample\n",
            id, a, d, s, r, sl, smpid,  id);
    }
    inline void ChannelFreq(unsigned ch, unsigned long freq)
    {
    	static unsigned long freqs[8];
    	if(freqs[ch] != freq)
		{
			freqs[ch] = freq;
			
			/* Frequency can change while playing because of vibrato */
			/* The first freq after "key on" is the note frequency */
			double time = GetClock(ch);
			fprintf(fp, "F%u@%g:%lu #freq\n", ch, time, freq);
		}
        FlushKon(ch);
    }
    inline void Volumes(unsigned ch, int left, int right)
    {
    	static int lvol[8], rvol[8];
    	if(lvol[ch]==left && rvol[ch]==right) return;
    	lvol[ch]=left; rvol[ch]=right;
    	
        /* Volume can change any time */
        double time = GetClock(ch);
        fprintf(fp, "V%u@%g:%d %d #volumes\n", ch, time, left, right);
    }
    inline void ChannelKeyOff(unsigned ch)
    {
        /* Enters the release phase */
        double time = GetClock(ch);
        fprintf(fp, "K%u@%g #koff\n", ch, time);
    }
    inline void ChannelKeyOn(unsigned ch)
    {
        S3Mkonpending[ch] = true;
		S3Mkontime[ch] = GetClock(ch);
    }
    inline void ChannelInsWas(unsigned ch, unsigned id)
    {
        /* Tells what instrument was it that was just PREVIOUSLY played */
        fprintf(fp, "W%u:%u #instrument was at %g\n", ch, id, GetClock(ch));
    }
    inline void Tick(unsigned ch, unsigned count)
    {
        S3Mclock[ch] += S3Mprevtick[ch];
        S3Mprevtick[ch] = count;
        S3Mclock2[ch] = 0;
    }
    inline void LocalTick(unsigned ch)
    {
        S3Mclock2[ch]++;
    }
    FILE *getfp() const { return fp; }
} S3Mmessage;

/*** CRC32 calculation (CRC::update) ***/
static class CRC
{
	unsigned long crctable[256];
public:
	CRC()
	{
		unsigned long poly = 0xEDB88320L;
		for(unsigned i=0; i<256; ++i)
		{
			unsigned long crc = i;
			for(unsigned j=8; j>0; --j) if(crc&1) crc = (crc>>1)^poly; else crc >>= 1;
			crctable[i] = crc;
		}
	}
	inline unsigned long update(unsigned long crc, unsigned char b) const
	{
		return ((crc >> 8) & 0x00FFFFFF) ^ crctable[(crc^b) & 0xFF];
	}
} CRC32;

typedef struct
{
	unsigned crc;
	unsigned uncompressedlength;
} SampleIdentification;

static void IdentifySample(SampleIdentification &id,
                           const signed char *compressed)
{
	unsigned long sum = 0xFFFFFFFF;
	unsigned length = SOUND_DECODE_LENGTH;
	const signed char *data = compressed;
    for(;;)
    {
    	unsigned char filter = *data;
    	bool lastblock = filter & 1;
    	
    	// filter &= ~2; // don't take loop flag to crcsum
    	
    	for(unsigned c=0; c<9; ++c)
    		sum = CRC32.update(sum, *data++);
		
    	if(lastblock)break;

		length += SOUND_DECODE_LENGTH;
    }
	id.crc                = sum;
	id.uncompressedlength = length;
}

struct PcmData
{
	vector<signed short> data;
	unsigned supposedlength;
	unsigned knownlength;
	
	PcmData(unsigned suplen) : data(suplen,0), supposedlength(suplen), knownlength(0)
	{
	}
	
	void Upgrade(unsigned pcmpos, const signed short *block)
	{
		if(knownlength > pcmpos)return;
		if(pcmpos + SOUND_DECODE_LENGTH > data.size())
			data.resize(supposedlength = pcmpos + SOUND_DECODE_LENGTH, 0);
		
		//for(unsigned a=0,p=pcmpos; a<SOUND_DECODE_LENGTH; )
		//	data[p++] = block[a++];
		memcpy(&data[pcmpos], &block[0], SOUND_DECODE_LENGTH * sizeof(*block));

	    knownlength += SOUND_DECODE_LENGTH;
	}
};
typedef map<unsigned, PcmData> pcms_t;
static pcms_t pcms;
typedef pcms_t::const_iterator pcms_cit;
typedef pcms_t::iterator pcms_it;

struct SamplePiece
{
	unsigned pcmid;
	unsigned count;
	inline bool operator== (const SamplePiece &b) const
	{ return pcmid == b.pcmid && count == b.count; }
	inline bool operator!= (const SamplePiece &b) const
	{ return pcmid != b.pcmid || count != b.count; }
};
class SampData
{
	vector<SamplePiece> data;
	
	const SampData slice(unsigned pos1, unsigned size) const
	{
		SampData tmp;
		unsigned pos2 = size + pos1;
		if(pos2 > data.size()) pos2 = data.size();
		tmp.data.assign(data.begin() + pos1, data.begin() + pos2);
		return tmp;
	}
	
public:
	typedef enum
	{
		mismatch,
		exactmatch,
		we_small,
		he_small			
	} matchtype;
	
	void Debug() const
	{
		FILE *fp = S3Mmessage.getfp();
		fprintf(fp, "# Sample data:");
	    for(unsigned a=0; a<data.size(); ++a)
		{
			fprintf(fp, " ");
			if(data[a].count!=1)fprintf(fp, "%u*", data[a].count);
			fprintf(fp, "%08X", data[a].pcmid);
		}
		fprintf(fp, "\n");
	}
   
	void Dump(unsigned id) const
    {
		unsigned superlooplen   = 1;
		unsigned superloopcount = 1;
		
		unsigned datasize = data.size();

		/* Task: Find a superloop. */
		unsigned maxsuperloopenddist = datasize / 2;
		for(unsigned superloopenddist = 0;
			         superloopenddist < maxsuperloopenddist;
			         ++superloopenddist)
		{
			unsigned maxsuperlooplen = (datasize - superloopenddist) / 2;
			unsigned minsuperlooplen = superloopenddist + 1;
			
			for(unsigned superlooplentest = maxsuperlooplen;
				         superlooplentest >= minsuperlooplen;
			             --superlooplentest)
			{
				// Test whether there's a loop of the specified kind in the end.
				unsigned looppos = datasize - superlooplentest - superloopenddist;
				
				unsigned foundcount = 0;
				for(foundcount=1; ; ++foundcount)
				{
					unsigned testpos = looppos - foundcount * superlooplentest;
					bool match = true;
					for(unsigned a=0; a<superlooplentest; ++a)
						if(data[testpos+a] != data[looppos+a])
						{ match = false; break; }
					if(!match) break;
				}
				if(foundcount > 1)
				{
					// found a superloop at the specified position!
					superlooplen   = superlooplentest;
					superloopcount = foundcount;
					if(superloopenddist > 0)
					{
						// but the loop isn't at the end!
						// check if the end part is a start of a new loop.
						SampData loopingpiece = slice(looppos, superlooplen);
						SampData endingpiece  = slice(looppos + superlooplen, datasize);
						matchtype m = endingpiece.Compare(loopingpiece);
						if(m != we_small)
						{
							// If the ending piece doesn't accept being
							// a child of the looping piece, don't accept.
							superlooplen   = 0;
							superloopcount = 0;
							continue;
						}
						// Okay so far.
					}
					
					// if the data is:      12345645645,
					// the program says:    12345 645*
					// but it's actually:   123 456* 45
					// How to fix:
					//  Check if *before* the loopstart is a piece from the end of the loop.
					//  If there is, step the loop backwards.
					//  This grows the tail, but it will be discarded later.
					// Result will be:
					unsigned superloopbegin = datasize - superlooplen * superloopcount - superloopenddist;
					while(superloopbegin > 0 &&
						  data[superloopbegin-1] == data[datasize - superloopenddist - 1])
					{
						++superloopenddist;
						--superloopbegin;
					}
					

					// Now discard the tail, if it exists.
					datasize -= superloopenddist;
					break;
				}
			}
		}
		unsigned superloopbegin = datasize - superlooplen * superloopcount;
		bool superloop = false;
	    if(superloopcount >= 1)
		{
			// Cut it to _one_ loop
			datasize = superloopbegin + superlooplen;
			superloop = true;
		}
		if(superlooplen==1)superloop = false;
		
		// At this point, "superloopcount" is meaningless.
		// superlooplen(unsigned) and superloop(bool) matter.
		
		unsigned lastsize = datasize == 0 ? 0 : data[datasize-1].count;
		if(!superloop)
		{
			// No superloop. Check if the last item has duplicates.
			if(datasize > 0 && lastsize > 1)
			{
				lastsize = 1;
				superloop = true;
				superlooplen = 1;
			}
		}

		FILE *sfp = S3Mmessage.getfp();
		
		fprintf(sfp, "# Sample %X: ", id);
		unsigned loopbegin = 0;
		vector<signed short> sample;
		for(unsigned a=0; a<datasize; ++a)
		{
			unsigned count = data[a].count;
			unsigned pcmid = data[a].pcmid;
			
			if(a+1 == datasize)count = lastsize;
			
			if(a == datasize-superlooplen)
			{
				fprintf(sfp, "[L]");
				loopbegin = sample.size();
			}
			if(count != 1)fprintf(sfp, "%u*", count);
			fprintf(sfp, "%08X ", pcmid);

			pcms_t::const_iterator i = pcms.find(pcmid);
			const PcmData &pcm = i->second;
			for(unsigned c=0; c<count; ++c)
			{
				unsigned pcmsize = pcm.knownlength;
				const signed short *pcmdata = &*pcm.data.begin();
				sample.insert(sample.end(), pcmdata, pcmdata+pcmsize);
			}
		}
		fprintf(sfp, "\n");
		
        S3Mmessage.SampleData(id, sample);
        S3Mmessage.SampleInfo(id, loopbegin, sample.size(), superloop ? 3 : 1);
    }
	// Update the sample:
	//   Add a pcm data block.
	void Add(unsigned pcmid)
	{
		if(data.size() > 0 && data[data.size()-1].pcmid == pcmid)
			++data[data.size()-1].count;
	    else
		{
			SamplePiece tmp;
			tmp.pcmid = pcmid; tmp.count = 1;
			data.push_back(tmp);
		}
	}
	matchtype Compare(const SampData &b) const
	{
		/* Compare this sample with the given sample.
		 * Possible results:
		 *   - We are inferior  (we_small)
		 *   - He's inferior    (he_small)
		 *   - He doesn't match (mismatch)
		 *   - Exact match      (exactmatch)
		 */
		for(unsigned a=0; a<data.size(); ++a)
		{
			// Did he run out of data while we still stand?
			if(a >= b.data.size()) return he_small;
			
			// He still stands.
			const SamplePiece &my = data[a];
			const SamplePiece &his = b.data[a];
			
			// But talks different language?
			if(my.pcmid != his.pcmid) return mismatch;
			
			// Damn. Okay.
			// But are we bigger?
			if(my.count > his.count)
			{
				// We are bigger... Does he care?
				if(a+1 == b.data.size()) return he_small;
				// He doesn't care!
				return mismatch;
			}
			
			// Is _he_ bigger?
			if(his.count > my.count)
			{
				// Do we care?
				if(a+1 == data.size()) return we_small;
				// He's talking weird things
				return mismatch;
			}
		}
		// We ran out of data. Did he?
		if(b.data.size() > data.size()) return we_small;
		// Friends!
		return exactmatch;
	}
};
typedef map<unsigned, SampData> samples_t;
static samples_t samples;
typedef samples_t::const_iterator samples_cit;
typedef samples_t::iterator samples_it;

struct InstrData
{
    /* Rates */
    int attack, decay, sustain, release;
    /* Levels */
    int sustainlevel;
    /* SampleID */
    unsigned sampleid;
    
    /* Purpose of this function:
	 *    Get an unique integer identifying this sample.
	 */
	unsigned getid() const
    {
        static map<string,unsigned> instr_idmap;
        
        // Create a string that identifies this instrument
        char Buf[128];
        sprintf(Buf, "%06d%06d%06d%06d%06d%08X",
            attack,decay,sustain,release,sustainlevel,sampleid);
        
        map<string,unsigned>::const_iterator i = instr_idmap.find(Buf);
        if(i == instr_idmap.end()) return instr_idmap[Buf] = instr_idmap.size();
        return i->second;
    }
    void Dump(unsigned insno) const
    {
        S3Mmessage.InstrInfo(insno, attack,decay,sustain,release,sustainlevel,sampleid);
    }
};
typedef map<unsigned, InstrData> instrus_t;
static instrus_t instrus;
typedef instrus_t::const_iterator instrus_cit;
typedef instrus_t::iterator instrus_it;

/* The purpose of this class is solely to produce
 * instrument information.
 *
 * During play it produces only InsWas -messages.
 */
static class ChanData
{
	SampleIdentification curid;
	PcmData *curpcm;
	unsigned pcmpos;
	SampData *cursamp;
	InstrData *curinst;
	
	/* Updates curpcm and pcmpos to collect the data for sample now playing */
	void FindSample(unsigned chno, bool isloop=false)
	{
		pcms_t::iterator j = pcms.find(curid.crc);
		if(j == pcms.end())
		{
			fprintf(stderr, "New sample! CRC32 %08X, length %u - channel %u%s\n",
				curid.crc, curid.uncompressedlength,
					chno, isloop?" (loop extension)" : "");

			PcmData tmp(curid.uncompressedlength);
			j = pcms.insert(make_pair(curid.crc, tmp)).first;
			//j = pcms.find(curid.crc);
		}
		
		curpcm = &j->second;
		pcmpos = 0;
		//fprintf(stderr, "curpcm is now %p\n", (const void *)curpcm);
		//fflush(stderr);
	}
public:
	ChanData() : curpcm(0), pcmpos(0), cursamp(0), curinst(0)
	{
	}
	
    /* Called at the beginning of a sample.
     * Now is a good chance to get the sample id.
     */
    void KeyOn(unsigned chno,
               int atk, int dec, int sus, int rel, int suslev,
               const signed char *data
              )
    {
    	//fprintf(stderr, "KeyOn(%u) called\n", chno);
    	
		IdentifySample(curid, data);
		FindSample(chno);
		
		cursamp = new SampData;
		
		curinst = new InstrData;
		curinst->attack = atk;
		curinst->decay  = dec;
		curinst->sustain = sus;
		curinst->release = rel;
		curinst->sustainlevel = suslev;
    }
    
    /* Called always after DecodeBlock() */
    void HaveData(unsigned chno, const signed short *block)
    {
		if(!curpcm)
		{/*
			fprintf(stderr,
				"S3MFUN Error: HaveData(%u) called before KeyOn() or LoopBegin()\n", chno);
			fflush(stderr);*/
			return;
		}
		curpcm->Upgrade(pcmpos, block);
		pcmpos += SOUND_DECODE_LENGTH;
    }

    /* Called always at the begin of a loop.
     * Now is a good chance to get the sample id.
     */
    void LoopBegin(unsigned chno, const signed char *data)
    {
		IdentifySample(curid, data);
		FindSample(chno, true);
    }
    
    /* Called always when last block of the current sample reached.
     * May be continued with a loop.
     * Now is a good chance to save the pcm data.
     */
    void SampleEnd(unsigned chno)
    {
		if(!curpcm)return;
		/* curpcm was a pointer to the map, so the data was
		 * automagically saved.
		 * We set curpcm to NULL so that we catch an exception
		 * if we happen to accidentally try to upgrade it.
		 */
		curpcm = NULL;
		/* Update the current sample */
		if(!cursamp)
	    	/*fprintf(stderr, "SampleEnd(%u) called after EndSample()\n", chno)*/;
		else
			cursamp->Add(curid.crc);
    }
    
    /* Called when the sample ends playing (the channel becomes quiet).
     * Now is a good chance to save the sample.
     */
    void EndSample(int chno)
    {
		SampleEnd(chno); // ENsure
    	if(!cursamp)
    	{
    		// Channel is already quiet
    		return;
    	}
		//cursamp->Debug();
		unsigned smpid = 0;
		/* Task: Find which sample matches this one.
		 *  1) Found superset: discard this one
		 *  2) Found inferior: upgrade it
		 *  3) Add it
		 */
		samples_t::iterator i;
		// Find out which sample we relate to.
		for(i=samples.begin(); i!=samples.end(); ++i)
		{
			SampData &testee = i->second;
			SampData::matchtype m = cursamp->Compare(testee);

			if(m == SampData::we_small)
			{
				// Discard what we got in favor for the
				// more complete sample
				smpid = i->first;
				break;
			}
		    if(m == SampData::exactmatch)
			{
				// Discard what we got, there's already
				// the exact same
				smpid = i->first;
				break;
			}
			if(m == SampData::he_small)
			{
				// Good news, update the existing sample.
				testee = *cursamp;
				smpid = i->first;
				break;
			}
		}
		if(!smpid)
		{
			// New sample. Note the new smpid must be >0,
			// or the above "if" won't work next time.
			smpid = samples.size() + 1;
			samples.insert(make_pair(smpid, *cursamp));
		}
		// We've done our work with the sample data
		delete cursamp;
		cursamp = NULL;
		
		// Now finish the instrument		
		curinst->sampleid = smpid;
		
		// Find out which instrument are we
		unsigned insid = curinst->getid();
		// Update that instrument
		instrus[insid] = *curinst;

		// We've done our work with the instrument
		delete curinst;
		curinst = NULL;
		
		// Only thing left is to update the log.
		S3Mmessage.ChannelInsWas(chno, insid);
		instrus[insid].Dump(insid);
		//samples[smpid].Debug();

		//fprintf(stderr, "End sample, smpid=%u, insid=%u\n", smpid, insid);
    }
} ChData[8];

void S9xSPClog_SaveSamples()
{
    for(unsigned a=0; a<8; ++a) ChData[a].EndSample(a);
    for(instrus_cit i=instrus.begin(); i!=instrus.end(); ++i) i->second.Dump(i->first);
    for(samples_cit i=samples.begin(); i!=samples.end(); ++i) i->second.Dump(i->first);
}
#endif

//#define SPChum if(chnum!=0)return;
#define SPChum

void S9xSPClog_KeyOff(unsigned chnum)
{
#ifdef SPClog
SPChum
	S3Mmessage.ChannelKeyOff(chnum);
#endif
}
void S9xSPClog_KeyOn(unsigned chnum)
{
#ifdef SPClog
SPChum
	S3Mmessage.ChannelKeyOn(chnum);
#endif
}
void S9xSPClog_Tick(unsigned chnum, unsigned samplecount)
{
#ifdef SPClog
SPChum
    S3Mmessage.Tick(chnum, samplecount);
#endif
}
void S9xSPClog_LocalTick(unsigned chnum, unsigned long freq)
{
#ifdef SPClog
SPChum
    S3Mmessage.LocalTick(chnum);
    S3Mmessage.ChannelFreq(chnum, freq);
#endif
}
void S9xSPClog_VolumeChange(unsigned chnum, int lvol, int rvol)
{
#ifdef SPClog
SPChum
	S3Mmessage.Volumes(chnum, lvol, rvol);
#endif
}
void S9xSPClog_HaveData(unsigned chnum, const signed short *block)
{
#ifdef SPClog
SPChum
	ChData[chnum].HaveData(chnum, block);
#endif
}
void S9xSPClog_EndData(unsigned chnum)
{
#ifdef SPClog
SPChum
	ChData[chnum].SampleEnd(chnum);
#endif
}
void S9xSPClog_LoopBegin(unsigned chnum, const signed char *data)
{
#ifdef SPClog
SPChum
	ChData[chnum].LoopBegin(chnum, data);
#endif
}
void S9xSPClog_EndSample(unsigned chnum)
{
#ifdef SPClog
SPChum
	ChData[chnum].EndSample(chnum);
#endif
}
void S9xSPClog_RegisterSample
  (unsigned chnum, const signed char *data,
   int atk, int dec, int sus, int rel, int suslev)
{
#ifdef SPClog
SPChum
	ChData[chnum].EndSample(chnum); // ensure this
    ChData[chnum].KeyOn(chnum, atk, dec, sus, rel, suslev, data);
#endif
}
